 /*
  * Routines for controlled evaluation of host names, user names, and so on.
  * They are, in fact, wrappers around the functions that are specific for
  * the sockets or TLI programming interfaces. The request_info and host_info
  * structures are used for result cacheing.
  * 
  * These routines allows us to postpone expensive operations until their
  * results are really needed. Examples are hostname lookups and double
  * checks, or username lookups. Information that cannot be retrieved is
  * given the value "unknown" ("paranoid" in case of hostname problems).
  * 
  * When ALWAYS_HOSTNAME is off, hostname lookup is done only when required by
  * tcpd paranoid mode, by access control patterns, or by %letter expansions.
  * 
  * When ALWAYS_RFC931 mode is off, user lookup is done only when required by
  * access control patterns or %letter expansions.
  * 
  * Author: Wietse Venema, Eindhoven University of Technology, The Netherlands.
  */

#ifndef lint
static char sccsid[] = "@(#) eval.c 1.3 95/01/30 19:51:45";
#endif

/* System libraries. */

#include <stdio.h>
#include <string.h>

/* Local stuff. */
#include "namespace.h"
#include "tcpd.h"
#include "internal.h"

 /*
  * When a string has the value STRING_UNKNOWN, it means: don't bother, I
  * tried to look up the data but it was unavailable for some reason. When a
  * host name has the value STRING_PARANOID it means there was a name/address
  * conflict.
  */
const char unknown[] = "unknown";    /* lookup failed */
const char paranoid[] = "paranoid";  /* hostname conflict */

/* eval_user - look up user name */

char *eval_user(struct request_info *request)
{
    if (request->user[0] == 0) {
	strcpy(request->user, unknown);
	if (request->sink == 0 && 
            ( SA(&request->client.au)->sa_family != ADDR_NOT_INITIALIZED ) && 
            ( SA(&request->server.au)->sa_family != ADDR_NOT_INITIALIZED ) ) 
                rfc931(&request->client.au, &request->server.au, request->user);
    }
    return (request->user);
}

/* eval_hostaddr - look up printable address */

char *eval_hostaddr(struct host_info *host)
{
    if (host->addr[0] == 0) {
	strcpy(host->addr, unknown);
	if (host->request->hostaddr != 0)
	    host->request->hostaddr(host);
    }
    return (host->addr);
}

/* eval_hostname - look up host name */

char *eval_hostname(struct host_info *host)
{
    if (host->name[0] == 0) {
	strcpy(host->name, unknown);
	if (host->request->hostname != 0)
	    host->request->hostname(host);
    }
    return (host->name);
}

/* eval_hostinfo - return string with host name (preferred) or address */

char *eval_hostinfo(struct host_info *host)
{
    char   *hostname;

#ifndef ALWAYS_HOSTNAME				/* no implicit host lookups */
    if (host->name[0] == 0)
	return (eval_hostaddr(host));
#endif
    hostname = eval_hostname(host);
    if (HOSTNAME_KNOWN(hostname)) {
	return (host->name);
    } else {
	return (eval_hostaddr(host));
    }
}

/* eval_client - return string with as much about the client as we know */

char *eval_client(struct request_info *request)
{
    char *hostinfo = eval_hostinfo(&request->client);

#ifndef ALWAYS_RFC931				/* no implicit user lookups */
    if (request->user[0] == 0)
	return (hostinfo);
#endif
    if (STR_NE(eval_user(request), unknown)) {
	snprintf(request->working_buf, sizeof(request->working_buf), "%s@%s", 
                 request->user, hostinfo);
	return (request->working_buf);
    } else {
	return (hostinfo);
    }
}

/* eval_server - return string with as much about the server as we know */

char *eval_server(struct request_info *request)
{
    char *host = eval_hostinfo(&request->server);
    char *daemon = eval_daemon(request);

    if (STR_NE(host, unknown)) {
	snprintf(request->working_buf, sizeof(request->working_buf), "%s@%s", 
                 daemon, host);
	return (request->working_buf);
    } else {
	return (daemon);
    }
}

