 /*
  * This module implements a simple access control language that is based on
  * host (or domain) names, NIS (host) netgroup names, IP addresses (or
  * network numbers) and daemon process names. When a match is found the
  * search is terminated, and depending on whether PROCESS_OPTIONS is defined,
  * a list of options is executed or an optional shell command is executed.
  * 
  * Host and user names are looked up on demand, provided that suitable endpoint
  * information is available as sockaddr_in structures or TLI netbufs. As a
  * side effect, the pattern matching process may change the contents of
  * request structure fields.
  * 
  * Diagnostics are reported through syslog(3).
  * 
  * Compile with -DNETGROUP if your library provides support for netgroups.
  * 
  * Author: Wietse Venema, Eindhoven University of Technology, The Netherlands.
  */

#ifndef lint
static char sccsid[] = "@(#) hosts_access.c 1.21 97/02/12 02:13:22";
#endif

/* System libraries. */

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <syslog.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#ifdef NETGROUP
#include <netdb.h>
#include <rpcsvc/ypclnt.h>
#endif

/* Local stuff. */
#include "namespace.h"
#include "tcpd.h"
#include "internal.h"

	
/* Delimiters for lists of daemons or clients. */

static const char sep[] = ", \t\r\n";

/* Constants to be used as return values for match functions. */

#define	YES		1
#define	NO		0
#define	ERROR		-1

 /*
  * These variables are globally visible so that they can be redirected in
  * verification mode.
  */

const char *hosts_allow_table = HOSTS_ALLOW;
const char *hosts_deny_table = HOSTS_DENY;
int         hosts_access_verbose = 2;

 /*
  * In a long-running process (3rd party daemon), we are not at liberty 
  * to just go away or use the twist option. All programs that are part 
  * of this suite explicitly set this to 0. 
  */

int resident = 1;		/* 0: no; +1: yes */

/* Forward declarations. */
static int table_match(const char *table, struct request_info *request);
static int list_match(char *list, struct request_info *request, 
	int (*match_fn)(char*, struct request_info *));
static int server_match(char *tok, struct request_info *request);
static int client_match(char *tok, struct request_info *request);
static int host_match(char *tok, struct host_info *host);
static int string_match(const char *tok, const char *string);
static int masked_match(const char *, const char *, const char *);
static void ipv6_mask(struct in6_addr *in6p, int maskbits);

/* Size of logical line buffer. */

#define	BUFLEN 2048

/* hosts_access - host access control facility */

int hosts_access( struct request_info *request)
{
    volatile int verdict;

    /*
     * If the (daemon, client) pair is matched by an entry in the file
     * /etc/hosts.allow, access is granted. Otherwise, if the (daemon,
     * client) pair is matched by an entry in the file /etc/hosts.deny,
     * access is denied. Otherwise, access is granted. A non-existent
     * access-control file is treated as an empty file.
     * 
     * After a rule has been matched, the optional language extensions may
     * decide to grant or refuse service anyway. Or, while a rule is being
     * processed, a serious error is found, and it seems better to play safe
     * and deny service. All this is done by jumping back into the
     * hosts_access() routine, bypassing the regular return from the
     * table_match() function calls below.
     */
    verdict = setjmp(tcpd_buf);
    if (verdict != 0)
	return (verdict == TCPD_PERMIT); /* We returned via a longjmp */
    if (table_match(hosts_allow_table, request) == YES)
	return YES;
    return (table_match(hosts_deny_table, request) == NO);
}

/* table_match - match table entries with (daemon, client) pair */

static int table_match(const char *table, struct request_info *request)
{
    FILE   *fp;
    char    sv_list[BUFLEN];		/* becomes list of daemons */
    char   *cl_list;			/* becomes list of clients */
    char   *sh_cmd = NULL;		/* becomes optional shell command */
    int     match = NO;
    struct tcpd_context saved_context;

    saved_context = tcpd_context;	/* stupid compilers */

    /*
     * Between the fopen() and fclose() calls, avoid jumps that may cause
     * file descriptor leaks.
     */

    if ((fp = fopen(table, "r")) != 0) {
	struct stat st;

	tcpd_context.file = table;
	tcpd_context.line = 0;
	/*
	 * Make sure this is a regular file
	 */
	if ( fstat( fileno(fp), &st) == 0) {
            if (!S_ISREG(st.st_mode)) 
		return ERROR;
	}
	else
	    return ERROR;

	while (match == NO && xgets(sv_list, sizeof(sv_list), fp) != 0) {
            int m1, m2;
	    if (sv_list[strlen(sv_list) - 1] != '\n') {
		tcpd_warn("missing newline or line too long");
		continue;
	    }
	    if (sv_list[0] == '#' || sv_list[strspn(sv_list, " \t\r\n")] == 0)
		continue;
	    if ((cl_list = split_at(sv_list, ':')) == 0) {
		tcpd_warn("missing \":\" separator");
		continue;
	    }
	    sh_cmd = split_at(skip_ipv6_addr(cl_list), ':');
	    m1 = list_match(sv_list, request, server_match);
	    m2 = list_match(cl_list, request, client_match);
	    if ((m1 == ERROR) || (m2 == ERROR))
	        match = ERROR;
	    else
		match = m1 && m2;
	}
	if (match == NO && ferror(fp))
            match = ERROR;
	(void) fclose(fp);
    } else if (errno != ENOENT) {
	tcpd_warn("cannot open %s: %m", table);
	match = ERROR;
    }
    if (match == YES) {
	if (hosts_access_verbose > 2) /* tcpdmatch calls this */
	    tcpd_warn("matched: %s line %d", 
                tcpd_context.file, tcpd_context.line);
	else if (hosts_access_verbose > 1) /* The normal call */
	    syslog(LOG_DEBUG, "matched: %s line %d",
                tcpd_context.file, tcpd_context.line);
	if (sh_cmd) {
#ifdef PROCESS_OPTIONS
	    process_options(sh_cmd, request);
#else
	    char    cmd[BUFSIZ];
	    shell_cmd(percent_x(cmd, sizeof(cmd), sh_cmd, request));
#endif
	}
    }
    tcpd_context = saved_context;
    return match;
}

/* list_match - match a request against a list of patterns with exceptions */

static int list_match(char *list, struct request_info *request, 
	int (*match_fn)())
{
    char   *tok;
    int match = NO;
    
    /*
     * Process tokens one at a time. We have exhausted all possible matches
     * when we reach an "EXCEPT" token or the end of the list. If we do find
     * a match, look for an "EXCEPT" list and recurse to determine whether
     * the match is affected by any exceptions.
     */

    for (tok = strtok(list, sep); tok != 0; tok = strtok((char *) 0, sep)) {
	if (STR_EQ(tok, "EXCEPT"))		/* EXCEPT: give up */
	    return NO;
	if (match_fn(tok, request) == YES) {	/* YES: look for exceptions */
	    while ((tok = strtok((char *) 0, sep)) && STR_NE(tok, "EXCEPT"))
		 /* VOID */ ;
	    if (tok == 0)
                match = YES;
	    else if ((match = list_match(NULL, request, match_fn)) != ERROR)
		match = !match;
	    break;
	}
    }
    return match;
}

/* server_match - match server information */

static int server_match(char *tok, struct request_info *request)
{
    char *host;

    if ((host = split_at(tok + 1, '@')) == 0) {	/* plain daemon */
	return string_match(tok, eval_daemon(request));
    } else {					/* daemon@host */
	if (string_match(tok, eval_daemon(request)))
            return host_match(host, &request->server);
        return NO;
    }
}

/* client_match - match client information */

static int client_match(char *tok, struct request_info *request)
{
    char *host;

    if ((host = split_at(tok + 1, '@')) == 0) {	/* plain host */
	return host_match(tok, &request->client);
    } else {					/* user@host */
        if (string_match(tok, eval_user(request)))
	    return host_match(host, &request->client);
        return NO;
    }
}

/* host_match - match host name and/or address against pattern */

static int host_match(char *tok, struct host_info *host)
{
    char *net_mask;

    /*
     * This code looks a little hairy because we want to avoid unnecessary
     * hostname lookups.
     * 
     * The KNOWN pattern requires that both address AND name be known; some
     * patterns are specific to host names or to host addresses; all other
     * patterns are satisfied when either the address OR the name match.
     */

    if (tok[0] == '@') {			/* netgroup: look it up */
#ifdef  NETGROUP
	static char *mydomain = 0;
	if (mydomain == 0)
	    yp_get_default_domain(&mydomain);
	return (innetgr(tok + 1, eval_hostname(host), (char *) 0, mydomain));
#else
	tcpd_warn("netgroup support is disabled");	/* not tcpd_jump() */
	return ERROR;
#endif
    } else if (STR_EQ(tok, "KNOWN")) {		/* check address and name */
	char   *name = eval_hostname(host);
	return (STR_NE(eval_hostaddr(host), unknown) && HOSTNAME_KNOWN(name));
    } else if (STR_EQ(tok, "LOCAL")) {		/* local: no dots in name */
	char   *name = eval_hostname(host);
	return (strchr(name, '.') == 0 && HOSTNAME_KNOWN(name));
    } else if (tok[0] == '[') {
	char *cbr, *slash;
	int mask = 128;
        struct in6_addr in6, hostin6;

	sock_hostname(host);
	
	/* make sure SA structure was initted */
	if ( (int)SA(&host->au)->sa_family == ADDR_NOT_INITIALIZED ) 
	   return ERROR;

	/* FIXME: should we test for IPv4 mapped IPv6 ? */
	
	if ( SA(&host->au)->sa_family != AF_INET6 )
	    return NO;
	
	if ((cbr = strchr(tok, ']')))
	   *cbr = 0;
	else {
	   tcpd_warn("No closing ] for IPv6 numeric address");
	   return ERROR;
	}

	if ((slash = strchr(tok, '/'))) {
            *slash = '\0';
	    errno = 0;
            mask = strtol(slash+1, NULL, 10);
            if (errno || mask < 0 || mask > 128) {
                tcpd_warn("bad IP6 prefix specification");
                return ERROR;
            }
	}
        hostin6 = SAIN6(&host->au)->sin6_addr;
        if (inet_pton(AF_INET6, tok+1, &in6) != 1) {
            tcpd_warn("bad IP6 address specification");
            return ERROR;
        }
        /*
	 * Zero the bits we're not interested in in both addresses
	 * then compare.  Note that we take a copy of the host info
	 * in that case.
	 */
        if (mask != 128) {
	    ipv6_mask(&in6, mask);
	    ipv6_mask(&hostin6, mask);
	}
        if (memcmp(&in6, &hostin6, sizeof(in6)) == 0)
            return YES;
        return NO;
    } else if ((net_mask = split_at(tok, '/')) != 0) {	/* IPv4 net/mask */
	return masked_match(tok, net_mask, eval_hostaddr(host));
    } else {					/* anything else */
        return (string_match(tok, eval_hostaddr(host)) ||
	        (NOT_INADDR(tok) && NOT_INADDR6(tok) && 
                 string_match(tok, eval_hostname(host))) );
    }
}

/* string_match - match string against pattern */

static int string_match(const char *tok, const char *string)
{
    int n;

    if (tok[0] == '.') {			/* suffix */
	n = strlen(string) - strlen(tok);
	return (n > 0 && STR_EQ(tok, string + n));
    } else if (STR_EQ(tok, "ALL")) {		/* all: match any */
	return YES;
    } else if (STR_EQ(tok, "KNOWN")) {		/* not unknown */
	return (STR_NE(string, unknown));
    } else if (tok[(n = strlen(tok)) - 1] == '.') {	/* prefix */
	return (STRN_EQ(tok, string, n));
    } else {					/* exact match */
	return (STR_EQ(tok, string));
    }
}

/* masked_match - match IPv4 address against netnumber/netmask */

static int masked_match(const char *net_tok, const char *mask_tok, 
	const char *string)
{
    int ret_val = NO;
    tcpd_sockaddr *addr;

    if ((addr = numeric_addr(string)) == NULL)
	return ret_val;
    
    if (SA(addr)->sa_family == AF_INET) {
        tcpd_sockaddr *net, *mask=NULL;
	uint32_t m;

	/* Check for the 255.255.255.0 style mask */
        if (strchr(mask_tok, '.') != NULL) {
            mask = numeric_addr(mask_tok);
            if ((mask == NULL) || (SA(mask)->sa_family != AF_INET)) {
	        tcpd_warn("Bad IPv4 network mask: %s", mask_tok);
		free(addr);
		if (mask) 
		    free(mask);
	        return ERROR;
	    }
	    m = (uint32_t)ntohl(SAIN(mask)->sin_addr.s_addr);
	    free(mask);
	}
	else /* OK, must be bit mask */
	{
            long tmp_m;
	    int i;
		    
	    errno = 0;
	    tmp_m = strtol(mask_tok, NULL, 10);
	    if (errno != 0) {
	        tcpd_warn("Bad IPv4 net mask: %s %m", mask_tok);
		free(addr);
	        return ERROR;
	    }
	    if (tmp_m > 32) {
	        tcpd_warn("IPv4 net mask > 32: %d", tmp_m);
		free(addr);
	        return ERROR;
	    }
	    m = 0;
	    for (i=0; i<tmp_m; i++) 
                m |= (uint32_t)(0x80000000UL >> i);
	}
	    
        net = numeric_addr(net_tok);
        if (net == NULL) {
	    tcpd_warn("Bad IPv4 network address: %s", net_tok);
	    ret_val = ERROR;
	} else {
	    uint32_t a, n;

	    a = (uint32_t)ntohl(SAIN(addr)->sin_addr.s_addr);
	    n = (uint32_t)ntohl(SAIN(net)->sin_addr.s_addr);
	    ret_val = ((a & m) == (n & m));
            free(net);
	}
    }
    free(addr);
    return ret_val; /* not tcpd_jump() */
}

/*
 * Function that zeros all but the first "maskbits" bits of the IPV6 address
 * This function can be made generic by specifying an address length as
 * extra parameter. (So Wietse can implement 1.2.3.4/16)
 */
static void ipv6_mask(struct in6_addr *in6p, int maskbits)
{
    unsigned char *p = (unsigned char*) in6p;
    int bytes, len;

    if (maskbits < 0 || maskbits >= 128)
        return;

    bytes = maskbits / 8;
    len = sizeof(struct in6_addr) - (bytes+1);
    maskbits %= 8;

    while (bytes) {
        *p++ &= 0xff;
	bytes--;
    }

    if (maskbits != 0)
        *p++ &= 0xff << (8 - maskbits);
    while (len--)
	*p++ = 0;
}

