 /*
  * @(#) internal.h 1.5 96/03/19 16:22:24
  * 
  * Author: Wietse Venema, Eindhoven University of Technology, The Netherlands.
  */

#ifndef IMPL_HEADER
#define IMPL_HEADER

#include <sys/socket.h>
#include <stdio.h>
#include <string.h>
#include <setjmp.h>
#include "tcpd.h"


/* Common socket casting operations */
#define SA( p )               ( (struct sockaddr *) (p) )
#define SAIN( p )             ( (struct sockaddr_in *) (p) )
#define SAIN6( p )            ( (struct sockaddr_in6 *) (p) )
#define ADDR_FAM( p )         ( (p)->sa.sa_family )

/* Common string operations. Less clutter should be more readable. */

#define STRN_CPY(d,s,l)	{ strncpy((d),(s),(l)); (d)[(l)-1] = 0; }
#define STRN_EQ(x,y,l)	(strncasecmp((x),(y),(l)) == 0)
#define STR_EQ(x,y)	(strcasecmp((x),(y)) == 0)
#define STR_NE(x,y)	(strcasecmp((x),(y)) != 0)

extern const char unknown[];
extern const char paranoid[];
extern const char *hosts_allow_table;   /* for verification mode redirection */
extern const char *hosts_deny_table;    /* for verification mode redirection */
extern int hosts_access_verbose;        /* for verbose matching mode */
extern int rfc931_timeout;              /* user lookup timeout */
extern int resident;                    /* > 0 if resident process */

/* Tests the results of evaluating the hostname. */
#define HOSTNAME_KNOWN(s) (STR_NE((s),unknown) && STR_NE((s),paranoid))

/* if the first mismatch isn't the null terminator */
#define NOT_INADDR(s) ((s)[strspn(s,"01234567890./")] != 0)

/* if there's no colon or the first mismatch isn't the null terminator */
#define NOT_INADDR6(s) ( (strchr(s, ':') == 0) || \
		        ((s)[strspn(s,"01234567890abcdef.:/")] != 0) )


 /*
  * While processing access control rules, error conditions are handled by
  * jumping back into the hosts_access() routine. This is cleaner than
  * checking the return value of each and every silly little function. The
  * (-1) returns are here because zero is already taken by longjmp().
  */

#define TCPD_PERMIT       1               /* permit access */
#define TCPD_DENY         (-1)            /* deny_access */
#define TCPD_ERROR        TCPD_DENY       /* XXX */

/*
 * The following define is used to test if the client or server address
 * union has been initialized. It occupies sa_family which gets set to
 * either AF_INET or AF_INET6.
 */

#define ADDR_NOT_INITIALIZED 255

/* Internal functions. */

extern void shell_cmd(const char *);	      /* execute shell command */
extern char *percent_x(char *, int, const char *, 
	struct request_info *);		      /* do %<char> expansion */
extern char *percent_m(char *, const char *); /* report errno where %m found */
extern void rfc931(const tcpd_sockaddr *, const tcpd_sockaddr *,
        char *dest);                      /* client name from RFC 931 daemon */
#ifdef __GNUC__
__attribute__ ((noreturn))
#endif
extern void clean_exit(const struct request_info *);/* clean up and exit */
extern char *xgets(char *, int, FILE *);            /* fgets() on steroids */
extern char *split_at(char *, int);                 /* strchr() and split */
extern tcpd_sockaddr *numeric_addr(const char *str);/* convert string to addr */
char *skip_ipv6_addr(char *str);                    /* ipv6 parser helper */
#ifdef KILL_IP_OPTIONS
void fix_options(struct request_info *);       /* kill IP options on a socket */
#endif
 
 /*
  * Routines for delayed evaluation of request attributes. Each attribute
  * type has its own access method. The trivial ones are implemented by
  * macros. The other ones are wrappers around the transport-specific host
  * name, address, and client user lookup methods. The request_info and
  * host_info structures serve as caches for the lookup results.
  */

extern char *eval_user(struct request_info *);	 /* client user */
extern char *eval_hostname(struct host_info *);	 /* printable hostname */
extern char *eval_hostaddr(struct host_info *);	 /* printable host address */
extern char *eval_hostinfo(struct host_info *);	 /* host name or address */
extern char *eval_client(struct request_info *); /* whatever is available */
extern char *eval_server(struct request_info *); /* whatever is available */

#define eval_daemon(r)	((r)->daemon)	/* daemon process name */
#define eval_pid(r)	((r)->pid)	/* process id */

/* Socket-specific methods, including DNS hostname lookups. */

#if defined(TLI) || defined(PTX) || defined(TLI_SEQUENT)
extern void tcpd_sock_host(struct request_info *);/* look up endpoint addr'es */
#endif
extern void sock_hostname(struct host_info *);/* xlate address to hostname */
extern void sock_hostaddr(struct host_info *);/* address to printable address */
#define sock_methods(r) \
	{ (r)->hostname = sock_hostname; (r)->hostaddr = sock_hostaddr; }

/* The System V Transport-Level Interface (TLI) interface. */

#if defined(TLI) || defined(PTX) || defined(TLI_SEQUENT)
extern void tli_host(struct request_info *);/* lookup endpoint addresses etc. */
#endif

 /*
  * Problem reporting interface. Additional file/line context is reported
  * when available.
  */
#ifdef __GNUC_MINOR__
#ifndef __GNUC_MINOR
#define __GNUC_MINOR __GNUC_MINOR__
#endif
#endif
#if ( ( __GNUC__ >= 2 ) && ( __GNUC_MINOR > 90 ) )
#define ATTR_CHECK( l, b)				\
        __attribute__ ((format (printf, l, b)))		
#else
#define ATTR_CHECK( l, b)				
#endif

/* report problem and proceed */
extern void tcpd_warn(const char *, ...)ATTR_CHECK(1,2);

/* report problem and jump */
#ifdef __GNUC__
__attribute__ ((noreturn))
#endif
extern void tcpd_jump(const char *, ...)ATTR_CHECK(1,2);

extern jmp_buf tcpd_buf;

struct tcpd_context {
    const char   *file;	                /* current file */
    int           line;                 /* current line */
};
extern struct tcpd_context tcpd_context;

 /*
  * In verification mode an option function should just say what it would do,
  * instead of really doing it. An option function that would not return
  * should clear the dry_run flag to inform the caller of this unusual
  * behavior.
  */

extern void process_options(char *, struct request_info *);/* execute options */
extern int dry_run;			/* verification flag */

/*
 * sleep() & alarm replacements
 */
void safe_sleep(int seconds);

/* Bug workarounds. */

#ifdef BROKEN_FGETS			/* partial reads from sockets */
#define fgets tcpd_fix_fgets
extern char *tcpd_fix_fgets(char *, int, FILE *);
#endif

#ifdef RECVFROM_BUG			/* no address family info */
#define recvfrom tcpd_fix_recvfrom
extern int tcpd_fix_recvfrom(int, char *, int, int, struct sockaddr *, int *);
#endif

#ifdef GETPEERNAME_BUG			/* claims success with UDP */
#define getpeername tcpd_fix_getpeername
extern int tcpd_fix_getpeername(int, struct sockaddr *, socklen_t *);
#endif

#ifdef USE_STRSEP			/* libc calls strtok() */
#define strtok	tcpd_fix_strtok
extern char *tcpd_fix_strtok(char *, const char *);
#endif

#ifdef LIBC_CALLS_STRTOK		/* libc calls strtok() */
#define strtok	tcpd_strtok
extern char *tcpd_strtok(char *, const char *);
#endif

#endif

