 /*
  * Misc routines that are used by tcpd and by tcpdchk.
  * 
  * Author: Wietse Venema, Eindhoven University of Technology, The Netherlands.
  */

#ifndef lint
static char sccsid[] = "@(#) misc.c 1.2 96/02/11 17:01:29";
#endif

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>

#include "namespace.h"
#include "tcpd.h"
#include "internal.h"


/* xgets - fgets() with backslash-newline stripping */

char *xgets(char *ptr, int len, FILE *fp)
{
    int     got;
    char   *start = ptr;

    while (fgets(ptr, len, fp)) {
	got = strlen(ptr);
	if (got >= 1 && ptr[got - 1] == '\n') {
	    tcpd_context.line++;
	    if (got >= 2 && ptr[got - 2] == '\\') {
		got -= 2;
	    } else {
		return (start);
	    }
	}
	ptr += got;
	len -= got;
	ptr[0] = 0;
    }
    if (ferror(fp))
	    tcpd_warn("Error reading table: %m");
    return (ptr > start ? start : 0);
}

/* split_at - break string at delimiter or return NULL */

char *split_at(char *string, int delimiter)
{
    char   *cp;

    if ((cp = strchr(string, delimiter)) != 0)
	*cp++ = 0;
    return (cp);
}

/* 
 * creates a tcpd_sockaddr from a given string. Caller must
 * free the data structure. At the moment only returns first,
 * if a dual stack machine with more than 1 address, could have
 * a problem.
 */
tcpd_sockaddr *numeric_addr(const char *str)
{
    tcpd_sockaddr *addr;
    struct addrinfo hints, *res;
      
    memset(&hints, 0, sizeof(hints));
    /*
     * Pick a socket type so multiple records don't come back. All we want
     * is the internal numeric address representation
     */
    hints.ai_socktype= SOCK_STREAM;	
    hints.ai_family = PF_UNSPEC;
    hints.ai_flags = AI_NUMERICHOST;
    if ( getaddrinfo(str, NULL, &hints, &res) ) 
        return NULL;
    addr = (tcpd_sockaddr *)malloc(sizeof(tcpd_sockaddr));
    if (addr == NULL)
        return NULL;
    
    memcpy(addr, res->ai_addr, res->ai_addrlen);
    freeaddrinfo(res);
    
    return addr;
}

/* 
 * safe_sleep - provides a sleep function that does not interact with 
 * signal handlers
 */
void safe_sleep(int seconds)
{
    struct timespec ts, left;
    ts.tv_sec = seconds;
    ts.tv_nsec = 0;
    left.tv_sec = 0;
    left.tv_nsec = 0;
    errno = 0;

    while (nanosleep(&ts, &left) == -1) {
        if (errno == EINTR)
            ts = left;
        else break;
    }
}

/*
 * When using IPv6 addresses, we'll be seeing lots of ":"s;
 * we require the addresses to be specified as [address].
 * An IPv6 address can be specified in 3 ways:
 *
 * x:x:x:x:x:x:x:x              (fully specified)
 * x::x:x:x:x                   (zeroes squashed)
 * ::FFFF:1.2.3.4               (IPv4 mapped)
 *
 * These need to be skipped to get at the ":" delimeters.
 *
 * We also allow a '/prefix' specifier.
 */

char *skip_ipv6_addr(char *str)
{
    char *obr, *cbr, *colon;
    char *q;

    while (1) 
    {
        if ((colon = strchr(str, ':')) == NULL)
	    return str;
	if ((obr = strchr(str, '[')) == NULL || obr > colon)
	    return str;
	if ((cbr = strchr(obr+1, ']')) == NULL)
	    return str;

	for (q = obr + 1; q < cbr; q++) 
	{
	    /*
             * Quick and dirty parse, cheaper than inet_pton
             * Could count colons and dots (must be 0 or 3 dots, no
             * colons after dots seens, only one double :, etc, etc)
             */
            if (*q != ':' && *q != '.' && *q != '/' && !isxdigit(*q & 0xff))
                return str;
        }
        str = cbr + 1;
    }
}

