 /*
  * rfc931() speaks a common subset of the RFC 931, AUTH, TAP, IDENT and RFC
  * 1413 protocols. It queries an RFC 931 etc. compatible daemon on a remote
  * host to look up the owner of a connection. The information should not be
  * used for authentication purposes. 
  * 
  * Diagnostics are reported through syslog(3).
  * 
  * Author: Wietse Venema, Eindhoven University of Technology, The Netherlands.
  */

#ifndef lint
static char sccsid[] = "@(#) rfc931.c 1.10 95/01/02 16:11:34";
#endif

/* System libraries. */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>

/* Local stuff. */
#include "namespace.h"
#include "tcpd.h"
#include "internal.h"

#define	RFC931_PORT	113		/* Semi-well-known port */
#define	ANY_PORT	0		/* Any old port will do */
#define RFC931_BUFSZ	512
#define SA_PORT(s) ADDR_FAM(s) == AF_INET ? \
	(s)->sa_in.sin_port : (s)->sa_in6.sin6_port

int rfc931_timeout = RFC931_TIMEOUT;
static int safe_select(int, fd_set *, fd_set *, fd_set *, struct timeval *);

/* fsocket - open stdio stream on top of socket */

static FILE *fsocket(int domain, int type, int protocol)
{
    int     s;
    FILE   *fp = NULL;

    if ((s = socket(domain, type, protocol)) < 0)
	tcpd_warn("socket: %m");
    else if ((fp = fdopen(s, "r+")) == 0) {
        tcpd_warn("fdopen: %m");
        close(s);
    }
    return (fp);
}

/* rfc931 - return remote user name, given socket structures */

void rfc931(const tcpd_sockaddr *rmt_sa, 
     const tcpd_sockaddr *our_sa, 
     char *dest)
{
    unsigned rmt_port;
    unsigned our_port;
    tcpd_sockaddr rmt_query;
    tcpd_sockaddr our_query;
    char   user[TCPD_STRING_LENGTH];
    char  *cp;
    const char  *result = unknown;
    FILE  *fp;
    int ret_val;

    /*
     * Use one unbuffered stdio stream for writing to and for reading from
     * the RFC931 etc. server. This is done because of a bug in the SunOS
     * 4.1.x stdio library. The bug may live in other stdio implementations,
     * too. When we use a single, buffered, bidirectional stdio stream ("r+"
     * or "w+" mode) we read our own output. Such behaviour would make sense
     * with resources that support random-access operations, but not with
     * sockets.
     */

    if ((fp = fsocket(ADDR_FAM(rmt_sa), SOCK_STREAM, 0)) != 0) {
        fd_set read_fds, write_fds;
        struct timeval tv;
        int flags;
        int sock = fileno(fp);	
        setbuf(fp, (char *) 0);

	/*
	 * Bind the local and remote ends of the query socket to the same
	 * IP addresses as the connection under investigation. We go
	 * through all this trouble because the local or remote system
	 * might have more than one network address. The RFC931 etc.
	 * client sends only port numbers; the server takes the IP
	 * addresses from the query socket.
	 */
	memcpy(&our_query, our_sa, sizeof(our_query));
	SA_PORT(&our_query) = htons(ANY_PORT);
	
	memcpy(&rmt_query, rmt_sa, sizeof(rmt_query));
	SA_PORT(&rmt_query) = htons(RFC931_PORT);

	do { 
	    ret_val = bind(sock, &our_query.sa, sizeof(our_query.sa));
	} while (ret_val < 0 && errno == EINTR);

	if (ret_val < 0 && errno != EINTR) {
            fclose(fp);
            tcpd_jump("bind: %m)");		    
        }

	/* Change to non-blocking */	
	flags = fcntl(sock, F_GETFL, 0);
	fcntl(sock, F_SETFL, flags | O_NONBLOCK);
	if ((ret_val = connect(sock, &rmt_query.sa,
		sizeof(rmt_query.sa))) < 0) {
            if (errno != EINPROGRESS)
		    tcpd_jump("non-blocking connect: %m");
	}
	
	/* Setup the timeout */
	FD_ZERO(&read_fds);
        FD_SET(sock, &read_fds);
        write_fds = read_fds;
	tv.tv_sec = rfc931_timeout;
	tv.tv_usec = 0;
	if ((ret_val = safe_select(sock+1, &read_fds, &write_fds,
			NULL, &tv)) <= 0) {
		close(sock);
		tcpd_jump("rfc931 timed out or errored connecting");
	}

	/* Now, see if the socket connected */
	if (FD_ISSET(sock, &read_fds) || FD_ISSET(sock, &write_fds)) {
            char buffer[RFC931_BUFSZ];
	    int len, error;

	    len = sizeof(error);
	    if (getsockopt(sock, SOL_SOCKET, SO_ERROR, &error, &len) < 0) {
		    close(sock);
		    tcpd_jump("getsockopt: %m");
	    }
	    if (error) {
		    close(sock);
		    errno = error;
		    tcpd_jump("Error connecting rfc931: %m");
	    }

	    /* No errors, set socket back to blocking */
	    fcntl(sock, F_SETFL, flags); 

            /*
             * Send query to server. Neglect the risk that a 13-byte
             * write would have to be fragmented by the local system and
             * cause trouble with buggy System V stdio libraries.
             */

            fprintf(fp, "%u,%u\r\n",
		ntohs(SA_PORT(rmt_sa)),
		ntohs(SA_PORT(our_sa)));
            fflush(fp);

            /*
             * Use asynchronous technique to timeout so we don't
             * hang forever waiting for response. Also clear the
	     * buffer for reading.
             */

            tv.tv_usec = 0;
            tv.tv_sec = rfc931_timeout;
	    memset(buffer, 0, RFC931_BUFSZ);
            ret_val = safe_select(sock+1, &read_fds, NULL, NULL, &tv);
            if (ret_val <= 0) {   
                /* Timed out or errored */
                fclose(fp);
                tcpd_jump("rfc931 timed out or errored reading");
            }

            /*
             * Read response from server. If the library incorrectly
             * assumes EOF when a read from a socket returns less than
             * requested, the compare will fail and we will ultimately deny 
	     * the connection since this is the safest thing to do.
             */
	    if (fread(buffer, sizeof(char), RFC931_BUFSZ-1, fp) != 0 && 
                ferror(fp) == 0 && 
		feof(fp) == 0 &&
	        sscanf(buffer, "%u , %u : USERID :%*[^:]:%255s",
		   &rmt_port, &our_port, user) == 3 &&
		ntohs(SA_PORT(rmt_sa)) == rmt_port &&
		ntohs(SA_PORT(our_sa)) == our_port) {

		/*
		 * Strip trailing carriage return. It is part of the
		 * protocol, not part of the data.
		 */
		if ( (cp = strchr(user, '\r')) )
                    *cp = 0;
		result = user;
	    }
	    else if (ferror(fp))
                tcpd_jump("Error %m reading rfc931 response");
	}
	fclose(fp);
    }
    STRN_CPY(dest, result, TCPD_STRING_LENGTH);
}

/*
 * safe_select - a wrapped version of select that handles EINTR.
 */

static int safe_select(int maxfds1, fd_set *readfds, fd_set *writefds,
                fd_set *exceptfds, struct timeval *tvptr)
{
    int ret_val;

#ifdef LINUX
    /*
     * Linux returns the unslept time in the timeval struct. We
     * simply recall select if we were interrupted.
     */
     do {
         ret_val = select(maxfds1, readfds, writefds, exceptfds, tvptr);
     } while (ret_val == -1 && errno == EINTR);

#else
    /*
     * Other OS's do not tell you how long you have been asleep.
     */

    time_t start, current;
    double expected = tvptr->tv_sec;

    time(&start);

    while ( (ret_val = select(maxfds1, readfds, writefds, exceptfds, tvptr) )
                == -1) {
        if (errno == EINTR) {
            double diff;
	    
            time(&current);
            diff = difftime(current, start);
            if (diff >= expected) {
                ret_val = 0;
                break;
            }
            else
                tvptr->tv_sec = expected - diff;
        }
    }
#endif
    return ret_val;
}

