 /*
  * This module determines the type of socket (datagram, stream), the client
  * socket address and port, the server socket address and port. In addition,
  * it provides methods to map a transport address to a printable host name
  * or address. Socket address information results are in static memory.
  * 
  * The result from the hostname lookup method is STRING_PARANOID when a host
  * pretends to have someone elses name, or when a host name is available but
  * could not be verified.
  * 
  * When lookup or conversion fails the result is set to STRING_UNKNOWN.
  * 
  * Diagnostics are reported through syslog(3).
  * 
  * Author: Wietse Venema, Eindhoven University of Technology, The Netherlands.
  */

#ifndef lint
static char sccsid[] = "@(#) socket.c 1.15 97/03/21 19:27:24";
#endif

/* System libraries. */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>

/* Local stuff. */
#include "namespace.h"
#include "tcpd.h"
#include "internal.h"

/* Forward declarations. */
static void sock_sink(int);

/* 
 * tcpd_sock_host - look up endpoint addresses and install conversion methods 
 */
void tcpd_sock_host(struct request_info *request)
{
    int     len;
    char    buf[BUFSIZ];
    int     fd = request->fd;

    sock_methods(request);

    /*
     * Look up the client host address. Hal R. Brand <BRAND@addvax.llnl.gov>
     * suggested how to get the client host info in case of UDP connections:
     * peek at the first message without actually looking at its contents. We
     * really should verify that client.sin_family gets the value AF_INET,
     * but this program has already caused too much grief on systems with
     * broken library code.
     */

    len = sizeof(struct sockaddr_in6);
    if (getpeername(fd, SA(&request->client.au), &len) < 0) {
	request->sink = sock_sink;
	len = sizeof(tcpd_sockaddr);
	if (recvfrom(fd, buf, sizeof(buf), MSG_PEEK,
		     SA(&request->client.au), &len) < 0) {
	    tcpd_warn("can't get client address: %m");
	    return;				/* give up */
	}
#ifdef really_paranoid
	memset(buf, 0 sizeof(buf));
#endif
    }

    /*
     * Determine the server binding. This is used for client username
     * lookups, and for access control rules that trigger on the server
     * address or name.
     */

    len = sizeof(tcpd_sockaddr);
    if (getsockname(fd, SA(&request->server.au), &len) < 0) {
	tcpd_warn("getsockname: %m");
	return;
    }
}

/*
 * sock_hostaddr - map endpoint address to printable form 
 */
void sock_hostaddr(struct host_info *host)
{
    const char *ret_val;
   
    /* Make sure the address union has been initialized */ 
    if ( ADDR_FAM( &host->au) == ADDR_NOT_INITIALIZED ) {
       host->addr[0] = 0;
       return;
    }

    if ( ADDR_FAM( &host->au ) == AF_INET )
        ret_val = inet_ntop(AF_INET, &SAIN(&host->au)->sin_addr,
			host->addr, sizeof(host->addr));
    else 
        ret_val = inet_ntop(AF_INET6, &SAIN6(&host->au)->sin6_addr,
			host->addr, sizeof(host->addr));

    if (ret_val == NULL)
        host->addr[0] = 0;
}

/* 
 * sock_hostname - map endpoint address to host name 
 */
void sock_hostname(struct host_info *host)
{
    tcpd_sockaddr *tsa = &host->au;
    socklen_t len;
   
    if ( tsa->sa.sa_family == ADDR_NOT_INITIALIZED )
        return;
    if ( tsa->sa.sa_family == AF_INET )
        len = sizeof(struct sockaddr_in);
    else if ( tsa->sa.sa_family == AF_INET6 )
        len = sizeof(struct sockaddr_in6);
    else
    {
        tcpd_warn("Unknown address family passed into sock_hostname");
        return;
    }
    /*
     * First, do a reverse lookup
     */
    if (getnameinfo( &tsa->sa, len, host->name, 
                     sizeof(host->name), 0, 0, NI_NAMEREQD | NI_NOFQDN) == 0)
    {
        struct addrinfo hints, *res, *ressave;
	int ret_val;

        memset(&hints, 0, sizeof(hints));
        hints.ai_flags = AI_CANONNAME;
        hints.ai_socktype = SOCK_STREAM;
        hints.ai_family = ADDR_FAM(tsa);
       	
        /*
	 * And then a forward lookup on what was returned.
	 */
        if( (ret_val = getaddrinfo(host->name, NULL, &hints, &res)) < 0 ) 
	{
	    tcpd_warn("can't verify hostname: getaddrinfo(%s) failed - %s",
		      host->name, gai_strerror(ret_val));
            strcpy(host->name, paranoid); /* name is bad, clobber it */
            return ;
        }

	ressave = res;
        if ( (res == NULL) || (res->ai_addr == NULL) ) 
	{
            tcpd_warn("can't verify hostname: no addresses returned");
            strcpy(host->name, paranoid); /* name is bad, clobber it */
            return;
        }

	/* 
	 * Verify also that getaddrinfo() and getnameinfo() return the same
	 * hostname, or rshd and rlogind may still end up being spoofed.
	 *
	 * On some systems, getaddrinfo might return  "localhost.domain".
	 *  This is a DNS artifact. We treat it  as a special case. 
	 *  When we can't believe the address list  from 
	 *  getnameinfo("localhost") we're in big trouble anyway.
	 */
	else if ( STR_NE(host->name, res->ai_canonname) &&
	          STR_NE(host->name, "localhost") && 
	          STR_NE(host->name, "ip6-localhost") )
	{
	    tcpd_warn("host name/name mismatch: %s != %s",
	        host->name, NI_MAXHOST, res->ai_canonname);
	}
	else
	{
	    /*
	     * Names are equal. Next verify that the address is a member of 
	     * the address list returned by getnameinfo(hostname).
	     */
	    while (res)
	    {
		if (tsa->sa.sa_family != res->ai_family)
			continue;
		if (res->ai_family == AF_INET)
	        {
		    if (memcmp(&SAIN(tsa)->sin_addr, 
			       &SAIN(res->ai_addr)->sin_addr,
			       sizeof(struct in_addr)) == 0)
	            {
                        freeaddrinfo(ressave);
                        return;		/* name is good, keep it */ 
	            }
		}
		else
		{
		    if (IN6_ARE_ADDR_EQUAL(&SAIN6(tsa)->sin6_addr, 
			      &SAIN6(res->ai_addr)->sin6_addr))
	            {
                        freeaddrinfo(ressave);
                        return;		/* name is good, keep it */
	            }
		}
	        res = res->ai_next;
	    }
	
            /*
             * The getaddrinfo() and getnameinfo() calls did not return
             * the same hostname. This could be a nameserver configuration
             * problem. It could also be that someone is trying to spoof us.
             */
	    tcpd_warn("host name/name mismatch2: %s != %.*s",
	        host->name, NI_MAXHOST, ressave->ai_canonname);
        }

        freeaddrinfo(ressave);
	strcpy(host->name, paranoid);		/* name is bad, clobber it */
    }
}

/* sock_sink - absorb unreceived IP datagram */

static void sock_sink(int fd)
{
    char    buf[BUFSIZ];
    tcpd_sockaddr tsin;
    socklen_t size = sizeof(tsin);

    /*
     * Eat up the not-yet received datagram. Some systems insist on a
     * non-zero source address argument in the recvfrom() call below.
     */
    (void) recvfrom(fd, buf, sizeof(buf), 0, &tsin.sa, &size);
}

