 /*
  * Routines for controlled update/initialization of request structures.
  * 
  * request_init() initializes its argument. Pointers and string-valued members
  * are initialized to zero, to indicate that no lookup has been attempted.
  * 
  * request_set() adds information to an already initialized request structure.
  * 
  * Both functions take a variable-length name-value list.
  * 
  * Diagnostics are reported through syslog(3).
  * 
  * Author: Wietse Venema, Eindhoven University of Technology, The Netherlands.
  */

#ifndef lint
static char sccsid[] = "@(#) update.c 1.1 94/12/28 17:42:56";
#endif

/* System libraries */

#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdlib.h> /* for free() */

/* Local stuff. */
#include "namespace.h"
#include "tcpd.h"
#include "internal.h"

/* request_fill - request update engine */

static struct request_info *request_fill(struct request_info *request, 
	va_list ap)
{
    int     key;
    unsigned int len = TCPD_STRING_LENGTH;
    char   *ptr;

    while ((key = va_arg(ap, int)) > 0) {
	switch (key) {
	default:
	    tcpd_warn("request_fill: invalid key: %d", key);
	    return (request);
	case RQ_FILE:
	    request->fd = va_arg(ap, int);
	    continue;

	    /*
	     * All other fields are strings with the same maximal length.
	     */

	case RQ_DAEMON:
	    ptr = request->daemon;
	    break;
	case RQ_USER:
	    ptr = request->user;
	    break;
	case RQ_CLIENT_NAME:
	    ptr = request->client.name;
	    len = NI_MAXHOST;
	    break;
	case RQ_CLIENT_ADDR:
	    ptr = request->client.addr;
	    len = INET6_ADDRSTRLEN;
	    break;
	case RQ_SERVER_NAME:
	    ptr = request->server.name;
	    len = NI_MAXHOST;
	    break;
	case RQ_SERVER_ADDR:
	    ptr = request->server.addr;
	    len = INET6_ADDRSTRLEN;
	    break;
	}
	STRN_CPY(ptr, va_arg(ap, char *), len);
    }
    return (request);
}

/* request_init - initialize request structure */

struct request_info *request_init( struct request_info *request, ...)
{
    struct request_info *r;
    va_list ap;

    /*
     * Initialize data members. We do not assign default function pointer
     * members, to avoid pulling in the whole socket module when it is not
     * really needed.
     */
    va_start( ap, request);
    memset((char *)request, 0, sizeof(struct request_info));
    request->fd = -1;
    strcpy(request->daemon, unknown);
    snprintf(request->pid, sizeof(request->pid), "%ld", (long)getpid());
    request->client.request = request;
    SA(&request->client.au)->sa_family = ADDR_NOT_INITIALIZED;
    request->server.request = request;
    SA(&request->server.au)->sa_family = ADDR_NOT_INITIALIZED;
    r = request_fill(request, ap);
    va_end( ap );
    return (r);
}

/* request_set - update request structure */

struct request_info *request_set( struct request_info *request, ...)
{
    struct request_info *r;
    va_list ap;

    va_start(ap, request);
    r = request_fill(request, ap);
    va_end(ap);
    return (r);
}

